from scipy.optimize import leastsq
import numpy as np
import pandas as pd


class ExponentialSignalAnalysis:

    def calculateCharacteristicTimeOfExponentialDecayInSecondsWithIntegral(exponentialDecayInSeconds, exponent):
        return exponent * np.trapz(exponentialDecayInSeconds, x=exponentialDecayInSeconds.index, axis=0) / (
                    exponentialDecayInSeconds.iloc[0] - exponentialDecayInSeconds.iloc[-1])

    def chargeApproach(sliceDF):
        exponent = 1
        sliceSquaredDF = sliceDF.pow(exponent)
        return ExponentialSignalAnalysis.calculateCharacteristicTimeOfExponentialDecayInSecondsWithIntegral(sliceSquaredDF, exponent)


    def energyApproach(sliceDF):
        exponent = 2
        sliceSquaredDF = sliceDF.pow(exponent)
        return ExponentialSignalAnalysis.calculateCharacteristicTimeOfExponentialDecayInSecondsWithIntegral(sliceSquaredDF, exponent)

    def linLog(myDF):
        right_exp = lambda p, x: -p[1] * x + p[0]
        err = lambda p, x, y: (right_exp(p, x) - y)

        time = myDF.index
        rate = np.log(myDF.values)
        v0 = [0., time.min()]
        out = leastsq(err, v0, args=(time, rate))
        v = out[0]  # fit parameters out []
        xxx = np.arange(min(time), max(time), time[1] - time[0])
        timeConstant = 1 / v[1]
        return timeConstant


    def expFit(myDF):
        right_exp = lambda p, x: p[0] * np.exp(-p[2] * (x - p[1]))
        err = lambda p, x, y: (right_exp(p, x) - y)

        time = myDF.index
        rate = myDF.values
        v0 = [0., 0., 0.]
        out = leastsq(err, v0, args=(time, rate))
        v = out[0]  # fit parameters out []
        xxx = np.arange(min(time), max(time), time[1] - time[0])
        ccc = right_exp(v, xxx)
        timeConstant = 1 / v[2]
        return timeConstant

    __nameToFunction = {"chargeApproach": chargeApproach,
                        "energyApproach": energyApproach,
                        "linReg": linLog,
                        "expFit": expFit}

    @staticmethod
    def getCharacteristicTimeOfExponentialDecay(df, functionNamesToApply):
        """

            **returns dataframe with the time constants, calculated with the chosen methods**

            Function returns a dataframe with time constant according to the input method chosen for the passed DataFrame object
            Note that passed DataFrame object should just contain the decay, the functions getWindowWithDecayOnly or
            getWindowWithGivenMeanAndStd can be used for filtering

            :type df: DF
            :type functionNamesToApply: list
            :return: DF

            - Example :

            from source.signals.SignalAnalysis import SignalAnalysis
            import pandas as pd
            import numpy as np

            x = np.arange(100)*0.001
            timeConstant = 0.8
            df = pd.DataFrame(data = np.exp(-x/timeConstant), index = x, columns = ["Voltage_1"])

            functionNamesToApply = ["chargeApproach", "energyApproach", "linReg", "eFit"]
            SignalAnalysis.getDecayTimeConstant(df, functionNamesToApply)

            >>> 					Voltage_1
            >>>chargeApproach		0.8
            >>>energyApproach		0.8
            >>>linReg				0.8
            >>>eFit					0.8

        """

        functionsToApply = [ExponentialSignalAnalysis.__nameToFunction[x] for x in functionNamesToApply]
        if isinstance(df, pd.DataFrame) and not df.empty:
            return df.aggregate(functionsToApply)
        else:
            return df

    @staticmethod
    def calculateChangeInCharacteristicOfExponentialDecay(originSeries, window=100):
        """Method returns the change in the characteristic of the exponential decay over time as a series"""
        # resampledSeries=SignalAnalysis.resampleSeries(originSeries)

        samplingTime = (originSeries.index[1] - originSeries.index[0])
        difSeries = originSeries.diff(periods=window) / samplingTime / window
        tauSeries = -originSeries.div(difSeries)
        return tauSeries