import unittest
from unittest import mock

import pandas as pd
from lhcsmapi.cals.CalsDbSignal import CalsDbSignal as CalsDbSignal
from lhcsmapi.Time import Time as tc


class TestCalsSignal(unittest.TestCase):

    def setUp(self):

        self.NAME = 'test'
        self.START = 0.0
        self.END = 3e-9
        self.DATA = [1.0, 1.2, 1.4]
        self.INDEX_SECS = [tc.convertToUnixTimeSeconds(i, 's') for i in [0, 1, 2]]
        self.INDEX_TIMESTAMPS = [tc.convertToPandasTimestamp(i, 's') for i in self.INDEX_SECS]

        self.mockLoggingDB = mock.Mock()
        self.defaultCALSSignal = CalsDbSignal(ldb=self.mockLoggingDB)
        self.df = pd.DataFrame(self.DATA, self.INDEX_TIMESTAMPS, columns=[self.NAME])
        self.calsDF = pd.DataFrame(self.DATA, self.INDEX_SECS, columns=[self.NAME])

    def tearDown(self):
        pass

    def assertEqualSignals(self, firstCALSSignal, secondCALSSignal):
        """Method assert each property of firstCALSSignal against the the same property of secondCALSSignal"""
        self.assertEqual(firstCALSSignal.Name, secondCALSSignal.Name)
        self.assertEqual(firstCALSSignal.StartTime, secondCALSSignal.StartTime)
        self.assertEqual(firstCALSSignal.EndTime, secondCALSSignal.EndTime)
        self.assertEqual(firstCALSSignal.Unit, secondCALSSignal.Unit)


    #region createSignal tests

    def test_createCALSSignal_noParameters_defaultCALSBehaviorCreated(self):
        """Tests that the default CalsSignal is created if no arguments are passed to the createSignalCALS method"""
        calsSignal = CalsDbSignal.buildSignal(ldb=self.mockLoggingDB, duration = [(0, 's')])
        self.assertEqualSignals(calsSignal, self.defaultCALSSignal)

    def test_createCALSSignal_startTimeAndEndTimeProvided_CorrectCALSBehaviorCreated(self):
        """Tests that the correct CalsSignal is created if start and end times are passed to the creaeSignalCALS method"""
        
        # assert:
        arbitraryCALSSignal = CalsDbSignal(["test"], 0.0, 1.0, ldb=self.mockLoggingDB) # common situation is name startTime and endTime

        # act:
        calsSignal = CalsDbSignal.buildSignal(name='test', startTime=0.0, endTime=(1.0, 's'), ldb=self.mockLoggingDB)

        # assert:
        self.assertEqualSignals(calsSignal, arbitraryCALSSignal)

    def test_createCALSSignal_startTimeAndSingleDurationComponentProvided_CorrectCALSBehavior(self):
        """Tests that the correct CalsSignal is created if start and end times are passed to the creaeSignalCALS method"""

        # assert:
        arbitraryCALSSignal = CalsDbSignal(["test"], 0.0, 1.0, ldb=self.mockLoggingDB)  # common situation is name startTime and endTime

        # act:
        calsSignal = CalsDbSignal.buildSignal(name='test', startTime=0, duration=(1.0, 's'), ldb=self.mockLoggingDB)

        # assert:
        self.assertEqualSignals(calsSignal, arbitraryCALSSignal)

    def test_createCALSSignal_startTimeAndTwoDurationComponentsProvided_CorrectCALSBehavior(self):
        """Tests that the correct CalsSignal is created if start and a single duration component are passed to the creaeSignalCALS method"""

        # assert:
        arbitraryCALSSignal = CalsDbSignal(["test"], 0.0, 2.0, ldb=self.mockLoggingDB)  # common situation is name startTime and endTime

        # act:
        calsSignal = CalsDbSignal.buildSignal(name='test', startTime=(1, 's'), duration=[(1.0, 's'), (1.0, 's')], ldb=self.mockLoggingDB)

        # assert:
        self.assertEqualSignals(calsSignal, arbitraryCALSSignal)

    def test_createCALSSignal_parameterIsInvalidType_TypeErrorIsThrown(self):
        """T[1.0, 1.2, 1.4]ests when a keyword that is of an invalid type, a TypeError is thrown"""

        with self.assertRaises(TypeError):
            CalsDbSignal.buildSignal(name=1, startTime=0, endTime=1)

    #endregion


    def test_getSignal_returnsCorrectDataFrame(self):
        """Tests that getSignal returns correct pandas DataFrame"""

        # arrange:
        queryResult = {self.NAME: [self.INDEX_SECS, self.DATA]}

        self.mockLoggingDB.get.return_value = queryResult

        uut = CalsDbSignal(self.NAME, self.START, self.END, ldb=self.mockLoggingDB)

        # act
        result = uut.getSignal(self.NAME, self.START, self.END, ldb=self.mockLoggingDB)

        # assert that pytimber LoggingDB's get method was called with the right arguments
        self.mockLoggingDB.get.assert_called_with(self.NAME, self.START, self.END)

        # assert that a correctly formatted DataFrame is returned
        pd.testing.assert_frame_equal(result, self.calsDF)


    def test_getSignalDF_correctDataFrame(self):
        """Tests that getSignalDF returns correct pandas DataFrame consisting of more than one signal"""

        # arrange:createSignalCALS
        columns = ['signal1', 'signal2', 'signal3']

        self.mockLoggingDB.get.side_effect = [{columns[0]: [self.INDEX_SECS, self.DATA]},
                                              {columns[1]: [self.INDEX_SECS, self.DATA]},
                                              {columns[2]: [self.INDEX_SECS, self.DATA]}]

        df = pd.DataFrame({columns[0]: self.DATA, columns[1]: self.DATA, columns[2]: self.DATA}, self.INDEX_SECS)

        uut = CalsDbSignal(columns, self.START, self.END, ldb=self.mockLoggingDB)

        # act
        result = uut.getSignalDataframe(columns, self.START, self.END, ldb=self.mockLoggingDB)

        # assert that pytimber LoggingDB's get method was called with the right arguments
        self.mockLoggingDB.get.assert_any_call(columns[0], self.START, self.END)
        self.mockLoggingDB.get.assert_any_call(columns[1], self.START, self.END)
        self.mockLoggingDB.get.assert_any_call(columns[2], self.START, self.END)

        # assert that a correctly formatted DataFrame is returned
        pd.testing.assert_frame_equal(result, df)

    def test_read_multipleNamesProvided_returnsCorrectDataFrameWithPandasTimestampsIndex(self):
        """Tests that read method correctly get a the correct DataFrame. That it then converts the index
        to pandas Timestamps, when the object is created with multiple signal names"""
        # arrange:
        columns = ['signal1', 'signal2', 'signal3']

        self.mockLoggingDB.get.side_effect = [{columns[0]: [self.INDEX_SECS, self.DATA]},
                                              {columns[1]: [self.INDEX_SECS, self.DATA]},
                                              {columns[2]: [self.INDEX_SECS, self.DATA]}]

        df = pd.DataFrame({columns[0]: self.DATA, columns[1]: self.DATA, columns[2]: self.DATA}, self.INDEX_SECS)

        uut = CalsDbSignal(columns, self.START, self.END, ldb=self.mockLoggingDB)

        # act
        result = uut.read()

        # assert that pytimber LoggingDB's get method was called with the right arguments
        self.mockLoggingDB.get.assert_any_call(columns[0], self.START, self.END)
        self.mockLoggingDB.get.assert_any_call(columns[1], self.START, self.END)
        self.mockLoggingDB.get.assert_any_call(columns[2], self.START, self.END)

        # assert that a correctly formatted DataFrame is returned
        pd.testing.assert_frame_equal(result, df)

    def test_read_singleNameProvided_returnsCorrectDataFrame(self):
        """Tests that read method correctly get a the correct DataFrame. That it then converts the index
        to pandas Timestamps, when the object is created with a single signal name"""

        self.mockLoggingDB.get.return_value = {self.NAME: [self.INDEX_SECS, self.DATA]}

        uut = CalsDbSignal([self.NAME], self.START, self.END, ldb=self.mockLoggingDB)

        # act
        result = uut.read()

        # assert that pytimber LoggingDB's get method was called with the right arguments
        self.mockLoggingDB.get.assert_called_with(self.NAME, self.START, self.END)

        # assert that a correctly formatted DataFrame is returned
        pd.testing.assert_frame_equal(result, self.calsDF)

    def test_getSignal_queryFailed_RaisesError(self):
        """Tests  that when a query fails a exception is thrown"""

        self.mockLoggingDB.get.side_effect = Exception
        uut = uut = CalsDbSignal([self.NAME], self.START, self.END, ldb=self.mockLoggingDB)

        with self.assertRaises(Exception):
            uut.getSignal(self.NAME)

